import numpy as np
import gurobipy
from gurobipy import *

###########NOTES###############
#THE FINCTIONS ARE DEFINED AS
    #overloaded_trans_vertices        ---> Minimizes Cadinality of Upper Average (CUA) as found in Section 4.3 of "Cardinality of Upper Average and Network Optimization"
    #binary_overloaded_trans_vertices ---> Minimizes Cardinality of Upper Tail (CUT) as found in Section 4.3 of "Cardinality of Upper Average and Network Optimization"

'''
THE SETUP IS AS FOLLOWS:
    
    demand_node_demands----> Vector of demands being sent to transhipment nodes.
    supply_node_limit------> Vector of supply for supply nodes
    num_transhipment_nodes---> number of transhipment nodes recieveing from supply node and sending to the demand nodes
    load_limit---------------> Threshold Paramter; or upper limit of load that can be put on any one transhipment node
    costs_S---------------> costs_S[i,j]:=cost of sending 1 unit of supply from supply node 'j' to transhipment node 'i'
    costs_D---------------> costs_D[i,j]:=cost of sending 1 unit of supply from transhipment node 'j' to demand node 'i'
    
    EXAMPLE of trivial setup where no nodes are overloaded: 
demand_node_demands = [10,10,10] -------> three demand nodes, with demand of 10 units each
supply_node_limit=[15,15]           -------> Two supply nodes with 15 units of supply to send to transhipment nodes
num_transhipment_nodes = 3       -------> three transhipment nodes that recieve from supply node and send to demand nodes
load_limit = 20                  -------> if a transhipment node is sending+recieving more than 20 units it is 'overloaded'
'''
def overloaded_trans_vertices(supply_node_limit, demand_node_demands, num_transhipment_nodes, load_limit, costs_S, costs_D):
    
    
    m = Model('Min CUA')
    m.params.Crossover=0
    T = num_transhipment_nodes
    L = load_limit
    num_supply = len(supply_node_limit)
    num_demand = len(demand_node_demands)


    # Variables
    z = []  
    for i in range(T):
        z.append(
            m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS, name="z" + str(i)))

    S = [] # S[i,j]==number of units that supply node 'j' sends to transhipment node 'i' 
    for i in range(T):
        
        S.append([m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS)
                  for j in range(num_supply)])

    D = [] # D[i,j]==number of units that demand node 'i' recieves from transhipment node 'j'
    for i in range(num_demand):
        D.append(
            [m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS) for j in range(T)])
    D=np.array(D)
    S=np.array(S)
    a = m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS, name="a")
    # Objective
    m.update()
    m.setObjective(quicksum(z), GRB.MINIMIZE)

    # Constraints
    for i in range(T):
        m.addConstr(z[i] >= quicksum(costs_S[i,j]*S[i,j] 
                             for j in range(num_supply)) + quicksum(costs_D[j,i]*D[j,i] for j in range(num_demand)) - a * L + 1)

    for i in range(num_demand):
        m.addConstr(quicksum(D[i,j]
                             for j in range(T)) ==  a*demand_node_demands[i])

    for j in range(num_supply):
        m.addConstr(quicksum(S[i,j]
                             for i in range(T)) ==  a*supply_node_limit[j])

    for i in range(T):
        m.addConstr( -quicksum(S[i,j] for j in range(num_supply)) + quicksum(D[j,i] for j in range(num_demand)) == 0 )
    m.optimize()

    a=a.X
    for i in range(D.shape[0]):
        for j in range(D.shape[1]):
            D[i,j]=D[i,j].X/float(a)
    for i in range(S.shape[0]):
        for j in range(S.shape[1]):
            S[i,j]=S[i,j].X/float(a)
    for i in range(T):
        z[i]=z[i].X



    return S,D,z,a


### MIP version of the transhipment problem ###
def binary_overloaded_trans_vertices(supply_node_limit, demand_node_demands, num_transhipment_nodes, load_limit, costs_S, costs_D):
    m = Model('Min CUT')

    T = num_transhipment_nodes
    L = load_limit
    num_supply = len(supply_node_limit)
    num_demand = len(demand_node_demands)


    # Variables
    E = []  # flow through edge (i,j)
    for i in range(T):
        E.append(
            m.addVar(obj=0, vtype=GRB.BINARY, name="E" + str(i)))

    S = []
    for j in range(T):
        S.append([m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS)
                  for i in range(num_supply)])

    D = []
    for i in range(num_demand):
        D.append(
            [m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS) for j in range(T)])
    D=np.array(D)
    S=np.array(S)
    a = m.addVar(lb=0, ub=GRB.INFINITY, obj=0, vtype=GRB.CONTINUOUS, name="a")
    
    # Objective
    m.update()
    m.setObjective(quicksum(E), GRB.MINIMIZE)

    # Constraints
    print costs_S.shape
    print costs_D.shape
    print T
    for i in range(T):
        m.addConstr(E[i] >=(
                            ( quicksum(costs_S[i,j]*S[i,j] for j in range(num_supply)) + quicksum(costs_D[j,i]*D[j,i] for j in range(num_demand)) 
                                                                                                                        )
                             - L ) / 10**5 )
                             

    for i in range(num_demand):
        m.addConstr(quicksum(D[i,j]
                             for j in range(T)) == demand_node_demands[i])

    for j in range(num_supply):
        m.addConstr(quicksum(S[i,j]
                             for i in range(T)) ==  supply_node_limit[j])
    for i in range(T):
        m.addConstr( -quicksum(S[i,j] for j in range(num_supply)) + quicksum(D[j,i] for j in range(num_demand)) ==0 )

    m.optimize()

    for i in range(D.shape[0]):
        for j in range(D.shape[1]):
            D[i,j]=D[i,j].X
    for i in range(S.shape[0]):
        for j in range(S.shape[1]):
            S[i,j]=S[i,j].X
    for i in range(T):
        E[i]=E[i].X
    



    return S,D,E
###############################################################################################

#demand node parameters
num_demand_nodes=10000
demand_node_demands = [10 for i in range(num_demand_nodes)]

#supply node parameters
supply_node_limit=[10*num_demand_nodes]
num_supply = len(supply_node_limit)

#transhipment node parameters
num_transhipment_nodes = 125
T = num_transhipment_nodes

# Generate edge costs uniformly on the interval (0,1)
costs_S=np.random.rand( T,num_supply  )
costs_D=np.random.rand( num_demand_nodes,T  )


#Run minimization of CUA
load_limit = .7*supply_node_limit[0]/float(num_transhipment_nodes) #set threshold paramter for CUA minimization

S, D, z, a = overloaded_trans_vertices(
                                       supply_node_limit, demand_node_demands, num_transhipment_nodes, load_limit, costs_S, costs_D)
                                       
#Run minimization of CUT
load_limit = .2*supply_node_limit[0]/float(num_transhipment_nodes) #set threshold paramter of CUT minimization


Sb, Db, Eb = binary_overloaded_trans_vertices(
    supply_node_limit, demand_node_demands, num_transhipment_nodes, load_limit, costs_S, costs_D)
    
